/*
 * tail-f-grep - `tail -f | grep' implementation by boost xpressive in C++
 *
 * Copyright (c) 2009 Taiji Yamada <taiji@aihara.co.jp>
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 */
#include <iostream>
#include <fstream>
#include <string>
#include <boost/lexical_cast.hpp>
#include <boost/xpressive/xpressive.hpp>

using namespace std;
using namespace boost;

class tail_f_grep {
private:
  string path, match, vmatch;
  unsigned int wait, nlines;
public:
  tail_f_grep(string _path, string _match = "", string _vmatch = "", unsigned int _wait = 1, unsigned int _nlines = 0)
    : path(_path), match(_match), vmatch(_vmatch), wait(_wait), nlines(_nlines)
  {
  }
  int loop()
  {
    ifstream in;
    streamoff prev_offset = -1, beg_offset, end_offset;
    string line;
    xpressive::sregex reg = xpressive::sregex::compile(match);
    xpressive::sregex vreg = xpressive::sregex::compile(vmatch);

    while (!0) {
      in.open(path.c_str());
      if (in.is_open()) {
	in.seekg(0, ios::beg);
	beg_offset = in.tellg();
	in.seekg(0, ios::end);
	end_offset = in.tellg();
	while (beg_offset < end_offset) {
	  in.seekg(--end_offset, ios::beg);
	  if (in.get() == '\n') { ++end_offset; break; }
	}
	if (prev_offset == -1 && nlines) {
	  unsigned int n = 0;

	  while (beg_offset < end_offset) {
	    in.seekg(--end_offset, ios::beg);
	    if (in.get() == '\n' && ++n > nlines) {
	      ++end_offset;
	      break;
	    }
	  }
	}
	if (prev_offset == -1)
	  prev_offset = end_offset;
	if (prev_offset > end_offset)
	  prev_offset = beg_offset;
	in.seekg(prev_offset, ios::beg);
	while (in.tellg() < end_offset && getline(in, line)) {
	  xpressive::smatch sm, vsm;

	  if (xpressive::regex_search(line, sm, reg) &&
	      (vmatch.empty()?!0:!xpressive::regex_search(line, vsm, vreg))) {
	    cout << line << endl;
	    cout.flush();
	  }
	}
	prev_offset = end_offset;
	in.close();
      }
      sleep(wait);
    }
    return 0;
  }
};

int usage(ostream &out, string me, int rc)
{
  out << "usage:\n\t" + me + " [-w wait] [-n nlines] [-v regex] regex filename" << endl;
  return rc;
}
int main(int argc, char *argv[])
{
  int a;
  unsigned int wait = 1, nlines = 10;
  string vmatch;

  for (a=1; a<argc; a++)
    if (argv[a] == string("--") && ++a<argc)
      break;
    else if (argv[a] == string("-h") || argv[a] == string("--help"))
      return usage(cout, argv[0], 0);
    else if (argv[a] == string("-w") && a+1<argc)
      wait = lexical_cast<unsigned int>(argv[++a]);
    else if (argv[a] == string("-n") && a+1<argc)
      nlines = lexical_cast<unsigned int>(argv[++a]);
    else if (argv[a] == string("-v") && a+1<argc)
      vmatch = argv[++a];
    else
      break;

  if (!(a+1<argc))
    return usage(cerr, argv[0], 1);

  tail_f_grep the_tail_f_grep(argv[a+1], argv[a], vmatch, wait, nlines);
  the_tail_f_grep.loop();
  return 0;
}
